/*--------------------------------------------------------------------------

   main.cpp v3.0 7/March/2008
   Copyright (C) 2008 J. Ribelles

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

  --------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <GL/glut.h>
#include <GL/glui.h>

using namespace std;

#include <list>

#include "main.h"
#include "object.h"

#ifndef M_PI
  #define M_PI 3.1415927
#endif


/* view frustrum */
GLdouble  aspect;        // aspect ratio  of the window

/* main variables */
int       main_win;      // id of the main window
Object    theObject;

/* rendering */
int showAxis=  1; // show the axis: X (red), Y (green), Z (blue)
int showFloor= 1; // show the plane XZ
float matrixTranslation[3]= {0.0, 0.0f, 0.0f};// translation
int f_count= 0, abortAnim= 0, anim= 0;
float alpha, x, y, z;

/* camera interaction */
float alfa= 15, beta= 15, radio= 25;
short mouseButtonPressed= NO_BUTTON_PRESSED;  // what button is pressed??
int   oldMouseX= 0, oldMouseY= 0;    // previous mouse coordinates

/* glui interface */
int primitive= 0; // the number of the primitive selected to be created
int transformation= 0; // the number of the transformation selected to be applied
GLUI *glui_subwin;                            // right menu: primitives 3D
GLUI *glui_subwin2;                           // left menu: transformations
GLUI *glui_subwin3=NULL;                           // load and save window
GLUI_EditText *text1, *text2, *text3, *text4; // edit text boxes for primitives
GLUI_EditText *text5, *text6, *text7, *text8; // edit text boxes for transf.
GLUI_Button *deletePrimitive, *save, *undo, *dob, *load;
GLUI_StaticText *ctm[4][4], *ntm[4][4];
int width= 2;  //width of the static text matrix labels
float value5, value6, value7, value8;
GLUI_Rollout *advanced;
GLUI_FileBrowser *fb;
void combinedMatrix (void);


/*--------------------------------------------------------------------------
   Convert polar to rectangular coordinates
  --------------------------------------------------------------------------*/
void convert (float alfa, float beta, float radio,
	      float *x, float *y, float *z) {

  float _alfa= alfa*M_PI/180.0, _beta= beta*M_PI/180.0;
  
  *z= radio*cos(_beta)*cos(_alfa);
  *x= radio*cos(_beta)*sin(_alfa);
  *y= radio*sin(_beta);
}

/*--------------------------------------------------------------------------
   Animation of the transformations... 
  --------------------------------------------------------------------------*/
void theTimerTransformation (int value) {

  float f= (f_count+1.0)/50.0;

  if (abortAnim) {
    f_count=   0;
    anim=      0;
    abortAnim= 0;
    return;
  }

  if ((f_count>0) || (value< 0)) theObject.eraseTransformation();
  
  switch (value) {
    // do ...
  case id_translate: theObject.addTransformation (id_translate, x*f, y*f, z*f, 0.0); break;
  case id_rotate:    theObject.addTransformation (id_rotate, x, y, z, alpha*f); break;
  case id_scale:     theObject.addTransformation (id_scale, (x-1)*f+1.0, (y-1)*f+1.0, (z-1)*f+1.0, 0.0); break;
    // undo ...
  case -id_translate:theObject.addTransformation (id_translate, x*(1.0-f), y*(1.0-f), z*(1.0-f), 0.0); break;
  case -id_rotate:   theObject.addTransformation (id_rotate, x, y, z, alpha*(1.0-f)); break;
  case -id_scale:    theObject.addTransformation (id_scale, (x-1)*(1-f)+1.0, (y-1)*(1-f)+1.0, (z-1)*(1-f)+1.0, 0.0); break;
  }
  
  if (f_count< 49) {
    f_count++;
    glutTimerFunc (40, theTimerTransformation, value);
  } else {
    f_count= 0;
    anim= 0;
    if (value< 0) theObject.eraseTransformation();
    dob->enable();
    if (theObject.emptyTransformations())
      undo->disable();
    else
      undo->enable();
    combinedMatrix(); 
    advanced->set_w(width);
  }
  
  if (glutGetWindow()!= main_win) 
    glutSetWindow(main_win);
  glutPostRedisplay();		  
}

/*--------------------------------------------------------------------------
   Finish the current animation of the transformations... 
  --------------------------------------------------------------------------*/
void finishAnimation (void) {

  if (anim== 2) 
    theObject.eraseTransformation(); 
  else {
    if (f_count> 0) theObject.eraseTransformation();
    theObject.addTransformation (transformation+1, x, y, z, alpha);
  }
  
  if (glutGetWindow()!= main_win) 
    glutSetWindow(main_win);
  glutPostRedisplay();		  
}

/*--------------------------------------------------------------------------
   Callback for making new primitives
  --------------------------------------------------------------------------*/
void cbPrimitives (int id) {

  float x1,x2,x3,x4;
  int i;

  switch (id) {
  case id_create: {
    if (anim) {
      finishAnimation();
      abortAnim= 1;
    }
    x1 = text1->get_float_val();
    x2 = text2->get_float_val();
    x3 = text3->get_float_val();
    x4 = text4->get_float_val();
    i = theObject.add (primitive+8, x1, x2, x3, x4);
    glui_subwin2->enable();
    save->enable();
    deletePrimitive->enable();
    undo->disable();
    combinedMatrix();
  }
    break;
  case id_delete: {
    if (anim) {
      abortAnim= 1;
    }
    theObject.erase();
    combinedMatrix();
    if (theObject.empty()) {
      save->disable();
      glui_subwin2->disable();
      deletePrimitive->disable();
    } else {
      dob->enable();
      if (theObject.emptyTransformations())
	undo->disable();
      else
	undo->enable();
    }
  }
    break;
  case id_radioPrimitives:
    switch (primitive+8) {
    case id_sphere:
      text1->enable();  text1->set_float_val(1.0);  text1->set_name("Radius");
      text2->enable();  text2->set_float_val(10.0); text2->set_name("Slices");
      text3->enable();  text3->set_float_val(10.0); text3->set_name("Stacks");
      text4->disable(); text4->set_float_val(0.0);  text4->set_name("");
      break;
    case id_cone:
      text1->enable();  text1->set_float_val(1.0);  text1->set_name("Base");
      text2->enable();  text2->set_float_val(1.0);  text2->set_name("Height");
      text3->enable();  text3->set_float_val(10.0); text3->set_name("Slices");
      text4->enable();  text4->set_float_val(10.0); text4->set_name("Stacks");
      break;
    case id_torus:
      text1->enable();  text1->set_float_val(1.0);  text1->set_name("InnerR");
      text2->enable();  text2->set_float_val(1.0);  text2->set_name("OuterR");
      text3->enable();  text3->set_float_val(10.0); text3->set_name("Slices");
      text4->enable();  text4->set_float_val(10.0); text4->set_name("Stacks");
      break;
    case id_cube:
    case id_teapot:
      text1->enable();  text1->set_float_val(1.0);  text1->set_name("Size");
      text2->disable(); text2->set_float_val(0.0);  text2->set_name("");
      text3->disable(); text3->set_float_val(0.0);  text3->set_name("");
      text4->disable(); text4->set_float_val(0.0);  text4->set_name("");  
      break;
    case id_dodecahedron:
    case id_octahedron:
    case id_tetrahedron:
    case id_icosahedron:
      text1->disable(); text1->set_float_val(0.0);  text1->set_name("");
      text2->disable(); text2->set_float_val(0.0);  text2->set_name("");
      text3->disable(); text3->set_float_val(0.0);  text3->set_name("");
      text4->disable(); text4->set_float_val(0.0);  text4->set_name("");  
      break;
    }
  }
  glutPostRedisplay();
}

void removeZeroDecimals (char *cadena) {
  
  int n= strlen(cadena), i, j= 0;
  
  for (i= n-1; i>= 0; i--)
    if (cadena[i]=='0') j++;
    else 
      if (cadena[i]== '.') { j++; break;}
      else 
	break;
  
  cadena[n-j]='\0';
}

/*---------------------------------------------------------------------
   Show the combined matrix of the selected object
  --------------------------------------------------------------------*/
void combinedMatrix (void){

  int i, j;
  char v[200];
  GLfloat m[16]={1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0};

  //  if ((theObject.empty()) || (f_count!= 0)) return;

  if ((!theObject.empty()) && (!theObject.emptyTransformations())) {
    glMatrixMode (GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    theObject.applyTransformations();
    glGetFloatv (GL_MODELVIEW_MATRIX, m);
    glPopMatrix();
  }

  for (i= 0; i< 4; i++)
    for (j= 0; j< 4; j++) {
      if ((m[i+j*4]<= 0.000001f) && (m[i+j*4]>= -0.000001f)) m[i+j*4]= 0.0f; 
      sprintf (v, "%.2f", m[i+j*4]);
      removeZeroDecimals(v);
      ntm[i][j]->set_text(v); 
      ntm[i][j]->set_w(width);
    }
}

/*---------------------------------------------------------------------
   Callback for the values of the transformations
  --------------------------------------------------------------------*/
void cbValueTransformation (int id){

  int i, j;
  char v5[200], v6[200], v7[200], v8[200];
  GLfloat m[16]={1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0};

  // identity
  for (i= 0; i< 4; i++)
    for (j= 0; j< 4; j++) {
      if (i == j) 
	ctm[i][j]->set_text("1");
      else 
	ctm[i][j]->set_text("0");
      ctm[i][j]->set_w(width);
    }
  
  sprintf (v5, "%.2f", value5);
  sprintf (v6, "%.2f", value6);
  sprintf (v7, "%.2f", value7);
  sprintf (v8, "%.2f", value8);

  removeZeroDecimals(v5);
  removeZeroDecimals(v6);
  removeZeroDecimals(v7);
  removeZeroDecimals(v8);

  switch (transformation+1) {
  case id_translate:
    ctm[0][3]->set_text(v5);
    ctm[1][3]->set_text(v6);
    ctm[2][3]->set_text(v7);
    break;
  case id_rotate: /* falta */
    glMatrixMode (GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glRotatef(value8, value5, value6, value7);
    glGetFloatv (GL_MODELVIEW_MATRIX, m);
    glPopMatrix();
    for (i= 0; i< 4; i++)
      for (j= 0; j< 4; j++) {
	if ((m[i+j*4]<= 0.000001f) && (m[i+j*4]>= -0.000001f)) m[i+j*4]= 0.0f; 
	sprintf (v5, "%.2f", m[i+j*4]);
	removeZeroDecimals(v5);
	ctm[i][j]->set_text(v5); 
	ctm[i][j]->set_w(width);
      }
    break;
  case id_scale:
    ctm[0][0]->set_text(v5);
    ctm[1][1]->set_text(v6);
    ctm[2][2]->set_text(v7);
    break;
  }
  
  advanced->set_w(width);
  glutPostRedisplay();
}

/*--------------------------------------------------------------------------
   Callback for transfromations options
  --------------------------------------------------------------------------*/
void cbTransformations (int id) {

  int type= 0;

  switch (id){
  case id_do:
    if (!theObject.empty()) {
      alpha = text8->get_float_val();
      x =     text5->get_float_val();
      y =     text6->get_float_val();
      z =     text7->get_float_val();
      anim=   1;
      dob->disable();
      undo->disable();
      glutTimerFunc (40, theTimerTransformation, transformation+1);
    }
    break;
  case id_undo:
    if (!theObject.empty())
      if (!theObject.emptyTransformations()) {
	anim= 2;
	dob->disable();
	undo->disable();
	theObject.getFirstTransformation (type, x, y, z, alpha);
	glutTimerFunc (40, theTimerTransformation, -type);
      }
    break;
  case id_radioTransformations:
    switch (transformation+1) {
    case id_translate:
      text5->enable();  text5->set_float_val(0.0);  text5->set_name("Tx");
      text6->enable();  text6->set_float_val(0.0);  text6->set_name("Ty");
      text7->enable();  text7->set_float_val(0.0);  text7->set_name("Tz");
      text8->disable(); text8->set_float_val(0.0);  text8->set_name("");
      break;
    case id_rotate:
      text5->enable();  text5->set_float_val(0.0);  text5->set_name("X");
      text6->enable();  text6->set_float_val(1.0);  text6->set_name("Y");
      text7->enable();  text7->set_float_val(0.0);  text7->set_name("Z");
      text8->enable();  text8->set_float_val(0.0);  text8->set_name("Angle");
      break;
    case id_scale:
      text5->enable();  text5->set_float_val(1.0);  text5->set_name("Sx");
      text6->enable();  text6->set_float_val(1.0);  text6->set_name("Sy");
      text7->enable();  text7->set_float_val(1.0);  text7->set_name("Sz");
      text8->disable(); text8->set_float_val(0.0);  text8->set_name("");
      break;
    }
    cbValueTransformation(0);
  }
}


void cbLoad( GLUI_Control* control )
{
  if (control->get_id() == id_load ) {
	if(glui_subwin3!=NULL) glui_subwin3->show();
	else {
    /****** Make command line window ******/
    glui_subwin3 = GLUI_Master.create_glui( "Load", 0, 50, 500 );
    //cmd_line = new GLUI_CommandLine( cmd_line_glui, "Command (try 'exit'):", NULL, -1, pointer_cb );
    //cmd_line->set_w( 400 );  /** Widen 'command line' control **/
    fb = new GLUI_FileBrowser(glui_subwin3, "Select a file", true, id_fb, cbLoad);

    GLUI_Panel *panel = new GLUI_Panel(glui_subwin3,"", GLUI_PANEL_NONE);
    //new GLUI_Button(panel, "Open", id_open, cbLoad);
    //new GLUI_Column(panel, false);
    new GLUI_Button(panel, "Cancel", id_cancel, cbLoad);
    
    glui_subwin3->set_main_gfx_window( main_win );
	}
    //control->disable();
  }
  else if ( control->get_id() == id_open ) {
	load->enable();
    control->glui->hide();
  }
  else if ( control->get_id() == id_cancel ) {
	load->enable();
    control->glui->hide();
    /*** User typed text into the 'command line' window ***/
    //printf( "Command (%d): %s\n", counter, cmd_line->get_text() );
    //std::string text = cmd_line->get_text();
    //if (text =="exit" || text == "quit")
//      exit(0);
  }
  else if ( control->get_id() == id_fb ) {
	std::string file_name;
    file_name = "";
    file_name = fb->get_file();
    if (theObject.load(file_name.c_str())) {
      if (anim) abortAnim= 1;
      combinedMatrix();
      if (!theObject.empty()) {
  	    deletePrimitive->enable();
	    save->enable();
	    glui_subwin2->enable();
	    if (theObject.emptyTransformations()) 
	      undo->disable();
	    else 
	      undo->enable();
      } else {
	    deletePrimitive->disable();
	    save->disable();
	    glui_subwin2->disable();
      }
    }
 	load->enable();
    control->glui->hide();
    glutPostRedisplay();
  } else {
 	load->enable();
    control->glui->hide();
  }
}

/*--------------------------------------------------------------------------
   Callback for file options
  --------------------------------------------------------------------------*/
void cbFile (int id) {

  switch (id){
  case id_save: //save 
    if (anim) {
      finishAnimation();
      abortAnim= 1;
      dob->enable();
      if (theObject.emptyTransformations())
	undo->disable();
      else
	undo->enable();
    }
    theObject.save();
    combinedMatrix();
    break;
  case id_load: //load
    if (theObject.load("data.save")) {
      if (anim) abortAnim= 1;
      combinedMatrix();
      if (!theObject.empty()) {
	deletePrimitive->enable();
	save->enable();
	glui_subwin2->enable();
	if (theObject.emptyTransformations()) 
	  undo->disable();
	else 
	  undo->enable();
      } else {
	deletePrimitive->disable();
	save->disable();
	glui_subwin2->disable();
      }
    }
    glutPostRedisplay();
    break;
  }
}

/*---------------------------------------------------------------------
   Creation of the interface for geometric primitives
  --------------------------------------------------------------------*/
void initInterfacePrimitives (void) {

  glui_subwin = GLUI_Master.create_glui_subwindow(main_win, GLUI_SUBWINDOW_RIGHT);
  GLUI_Panel *panelPrimitives = glui_subwin->add_panel("3D Primitives",true);
  GLUI_RadioGroup *radioPrimitives= glui_subwin->add_radiogroup_to_panel(panelPrimitives, &primitive, id_radioPrimitives, cbPrimitives);
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Sphere");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Cube");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Cone");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Torus");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Teapot");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Dodecahedron");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Octahedron");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Tetrahedron");
  glui_subwin->add_radiobutton_to_group (radioPrimitives, "Icosahedron");

  text1= glui_subwin->add_edittext_to_panel(panelPrimitives, "Radius", GLUI_EDITTEXT_FLOAT);
  text2= glui_subwin->add_edittext_to_panel(panelPrimitives, "Slices", GLUI_EDITTEXT_FLOAT);
  text3= glui_subwin->add_edittext_to_panel(panelPrimitives, "Stacks", GLUI_EDITTEXT_FLOAT);
  text4= glui_subwin->add_edittext_to_panel(panelPrimitives,       "", GLUI_EDITTEXT_FLOAT);
  text4->disable();

  text1->set_w(5);
  text2->set_w(5);
  text3->set_w(5);
  text4->set_w(5);
  
  text1->set_float_val(1.0);
  text2->set_float_val(10.0);
  text3->set_float_val(10.0);
  
  glui_subwin->add_button_to_panel(panelPrimitives, "Create", id_create, cbPrimitives);
  glui_subwin->add_separator_to_panel (panelPrimitives);
  deletePrimitive= glui_subwin->add_button_to_panel(panelPrimitives, "Delete", id_delete, cbPrimitives);
  deletePrimitive->disable();

  // File options
  GLUI_Panel *panelFile= glui_subwin->add_panel("File", true);
  load= glui_subwin->add_button_to_panel(panelFile, "Load", id_load, cbFile);
  save= glui_subwin->add_button_to_panel(panelFile, "Save", id_save, cbFile);
  save->disable();
  glui_subwin->add_separator_to_panel (panelFile);
  glui_subwin->add_button_to_panel(panelFile, "Quit", id_quit, (GLUI_Update_CB) exit);

  glui_subwin->set_main_gfx_window (main_win);
}

/*---------------------------------------------------------------------
   Creation of the interface for transformations
  --------------------------------------------------------------------*/
void initInterfaceTransformations (void) {

  int i, j;
  GLUI_Panel *currTM;

  glui_subwin2 = GLUI_Master.create_glui_subwindow(main_win, GLUI_SUBWINDOW_LEFT);
  glui_subwin2->disable();

  GLUI_Panel *panelTransformation = glui_subwin2->add_panel("Transformations", true);
  GLUI_RadioGroup *radioTransformations= glui_subwin2->add_radiogroup_to_panel(panelTransformation, &transformation, id_radioTransformations, cbTransformations);
  glui_subwin2->add_radiobutton_to_group (radioTransformations, "Translate");
  glui_subwin2->add_radiobutton_to_group (radioTransformations, "Rotate");
  glui_subwin2->add_radiobutton_to_group (radioTransformations, "Scale");

  text5= glui_subwin2->add_edittext_to_panel(panelTransformation, "Tx", GLUI_EDITTEXT_FLOAT, (void *)&value5, id_value5, cbValueTransformation);
  text6= glui_subwin2->add_edittext_to_panel(panelTransformation, "Ty", GLUI_EDITTEXT_FLOAT, (void *)&value6, id_value6, cbValueTransformation);
  text7= glui_subwin2->add_edittext_to_panel(panelTransformation, "Tz", GLUI_EDITTEXT_FLOAT, (void *)&value7, id_value7, cbValueTransformation);
  text8= glui_subwin2->add_edittext_to_panel(panelTransformation,   "", GLUI_EDITTEXT_FLOAT, (void *)&value8, id_value8, cbValueTransformation);
  text8->disable();

  text5->set_w(5);
  text6->set_w(5);
  text7->set_w(5);
  text8->set_w(5);

  dob=  glui_subwin2->add_button_to_panel(panelTransformation,"Do", id_do, cbTransformations);
  undo= glui_subwin2->add_button_to_panel(panelTransformation, "Undo", id_undo, cbTransformations);
  undo->disable();
  glui_subwin2->add_separator_to_panel(panelTransformation);
  advanced= glui_subwin2->add_rollout_to_panel(panelTransformation,"Matrices", false);

  currTM= glui_subwin2->add_panel_to_panel(advanced, "New", GLUI_PANEL_EMBOSSED);
  currTM->set_alignment(GLUI_ALIGN_CENTER);
  currTM->enable();
  for (j= 0; j< 4; j++) {
    for (i= 0; i< 4; i++) {
      if (i == j) 
	ctm[i][j]= glui_subwin2->add_statictext_to_panel(currTM, "1");
      else 
	ctm[i][j]= glui_subwin2->add_statictext_to_panel(currTM, "0");
      ctm[i][j]->set_w(width);
      ctm[i][j]->set_alignment(GLUI_ALIGN_CENTER);
      ctm[i][j]->enable();
    }
    if (j< 3) glui_subwin2->add_column_to_panel(currTM, false);
  }

  currTM= glui_subwin2->add_panel_to_panel(advanced, "Composed", GLUI_PANEL_EMBOSSED);
  currTM->set_alignment(GLUI_ALIGN_CENTER);
  currTM->enable();
  for (j= 0; j< 4; j++) {
    for (i= 0; i< 4; i++) {
      if (i == j) 
	ntm[i][j]= glui_subwin2->add_statictext_to_panel(currTM, "1");
      else 
	ntm[i][j]= glui_subwin2->add_statictext_to_panel(currTM, "0");
      ntm[i][j]->set_w(width);
      ntm[i][j]->set_alignment(GLUI_ALIGN_CENTER);
      ntm[i][j]->enable();
    }
    if (j< 3) glui_subwin2->add_column_to_panel(currTM, false);
  }

  glui_subwin2->set_main_gfx_window (main_win);
}

/*--------------------------------------------------------------------------
   Selecting a primitive for edition options
  --------------------------------------------------------------------------*/
void selectObject (int x, int y) {

  GLuint selectBuf[5];
  GLint  hits;
  GLint  viewport[4];
  float  _x, _y, _z;

  if (theObject.empty()) return;

  if (anim) {
    finishAnimation();
    abortAnim= 1;
    dob->enable();
    if (theObject.emptyTransformations())
      undo->disable();
    else
      undo->enable();
  }
  
  glGetIntegerv (GL_VIEWPORT, viewport);

  glSelectBuffer (5, selectBuf);
  glRenderMode (GL_SELECT);
  glInitNames();
  glPushName(0);

  glMatrixMode (GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluPickMatrix ((GLdouble) x, (GLdouble) (viewport[3]- y), 5.0, 5.0, viewport);
  gluPerspective (30.0, aspect, 0.05, 1000.0);

  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity ();
  glTranslatef (matrixTranslation[0]/10.0,matrixTranslation[1]/10.0, 0.0f);
  convert (alfa, beta, radio, &_x, &_y, &_z);
  gluLookAt (_x, _y, _z, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
  theObject.drawSelect();
  glMatrixMode (GL_MODELVIEW);
  glPopMatrix();
  glMatrixMode (GL_PROJECTION);
  glPopMatrix();

  hits= glRenderMode (GL_RENDER);
  if (hits> 0) {
    GLuint *ptr= selectBuf;
    ptr+= 3;
    //printf ("HITS: %d First: %d\n", hits, (GLuint)(*ptr));
    //		printf("selection\n"); 
    theObject.setCurrent((GLuint)(*ptr));
    if (theObject.emptyTransformations())
      undo->disable();
    else
      undo->enable();
    combinedMatrix();
  }
}

/*--------------------------------------------------------------------------
   Draw XYZ Axis
  --------------------------------------------------------------------------*/
void drawAxis (void) {

  glDisable (GL_LIGHTING);
  glLineWidth(2.0f);
  glBegin(GL_LINES);
    glColor3f(1.0f,0.0f,0.0f);
    glVertex3i (-100,0,0); // Red X
    glVertex3i ( 100,0,0);
    glColor3f(0.0f,1.0f,0.0f);
    glVertex3i (0,-100,0); // Green Y
    glVertex3i (0, 100,0);
    glColor3f(0.0f,0.0f,1.0f);
    glVertex3i (0,0,-100); // Blue Z
    glVertex3i (0,0, 100);
  glEnd();
  glLineWidth(1.0f);
  glEnable (GL_LIGHTING);
}

/*--------------------------------------------------------------------------
   Draw the floor
  --------------------------------------------------------------------------*/
void drawFloor (void) {

  int i, j;

  glColor3f (.75f, .75f, .75f);
  for (i= -100; i< 100; i+=1) {
    if ((showAxis==0) || (i!= 0)) {
      glBegin (GL_LINE_STRIP);
      for (j= -100; j< 100; j+=1) {
	glNormal3f (0,1,0);
	glVertex3i (i, 0, j);
      }
      glEnd();
    }
  }
  
  for (i= -100; i< 100; i+=1) {
    if ((showAxis==0) || (i!= 0)) {
      glBegin (GL_LINE_STRIP);
      for (j= -100; j< 100; j+=1) {
	glNormal3f (0,1,0);
	glVertex3i (j, 0, i);
      }
      glEnd();
    }
  }
}

/*--------------------------------------------------------------------------
   Drawing function
  --------------------------------------------------------------------------*/
void  display(void) {

  float x= 0, y= 0, z= 0; // camera position

  glClear (GL_COLOR_BUFFER_BIT| GL_DEPTH_BUFFER_BIT);
  
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (matrixTranslation[0]/10.0,matrixTranslation[1]/10.0, 0.0f);
  convert (alfa, beta, radio, &x, &y, &z);
  gluLookAt (x, y, z, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

  if (showFloor) drawFloor();
  if (showAxis)  drawAxis();

  glColor3f (0.95f, 0.95f, 0.95f);
  theObject.draw();

  glutSwapBuffers ();
}

/*--------------------------------------------------------------------------
   Render options...
  --------------------------------------------------------------------------*/
void initRenderOptions (void) {

  GLfloat light_direction[]= {3.0f, 2.0f, 5.0f, 1.0f};
  GLfloat l_diffuse[]=       {0.6f, 0.6f, 0.6f, 1.0f};
  GLfloat l_specular[]=      {0.1f, 0.1f, 0.1f, 1.0f};
  GLfloat lmodel_ambient[]=  {0.4f, 0.4f, 0.4f, 1.0f};

  GLfloat mat_diffuse[]=    {0.7f, 0.7f, 0.7f, 1.0f};
  GLfloat mat_specular[]=   {0.2f, 0.2f, 0.2f, 1.0f};
  GLfloat mat_shininess[]=  {50.0f};

  glDisable       (GL_CULL_FACE);
  glDepthFunc     (GL_LEQUAL);
  glEnable        (GL_DEPTH_TEST);
  glEnable        (GL_NORMALIZE);
  glShadeModel    (GL_FLAT);
  glPolygonMode   (GL_FRONT_AND_BACK, GL_FILL);
  glClearColor    (0.0f, 0.0f, 0.0f, 0.0f);

  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
//   glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE); comentted because teapot is not well shaded

  glLightfv (GL_LIGHT0, GL_POSITION, light_direction);
  glLightfv (GL_LIGHT0, GL_DIFFUSE,  l_diffuse);
  glLightfv (GL_LIGHT0, GL_SPECULAR, l_specular);
  glLightf  (GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.2);
  glLightf  (GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0);
  glLightf  (GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0);
  glEnable  (GL_LIGHTING);
  glEnable  (GL_LIGHT0);

  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_diffuse);
  glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
  glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
  glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glEnable (GL_COLOR_MATERIAL);
}

/*--------------------------------------------------------------------------
   View...
  --------------------------------------------------------------------------*/
void initView (void) {

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
}

/*--------------------------------------------------------------------------
   Attending the popup menu
  --------------------------------------------------------------------------*/
void menu (int option) {

  switch (option) {
  case 0: glShadeModel  (GL_SMOOTH); break;
  case 1: glShadeModel  (GL_FLAT);   break;
  case 4: glPolygonMode (GL_FRONT_AND_BACK, GL_FILL); break;
  case 5: glPolygonMode (GL_FRONT_AND_BACK, GL_LINE); break;
  case 2: showAxis=  1; break;
  case 3: showAxis=  0; break;
  case 6: showFloor= 1; break;
  case 7: showFloor= 0; break;
  case 11: glClearColor (0.4f, 0.4f, 0.5f, 0.0f); break;  // Blue
  case 12: glClearColor (0.0f, 0.0f, 0.0f, 0.0f); break;  // Black
  case 13: glClearColor (1.0f, 1.0f, 1.0f, 0.0f); break;  // White  }
  }
  glutPostRedisplay ();
}

/*--------------------------------------------------------------------------
   Defining the popup menu
  --------------------------------------------------------------------------*/
void initMenu (void) {

  int idMenu, idPMMenu /*, id2PMMenu*/;

  idMenu= glutCreateMenu (menu);
  glutAttachMenu (GLUT_RIGHT_BUTTON);

  idPMMenu= glutCreateMenu (menu);
  glutAddMenuEntry ("Yes", 2);
  glutAddMenuEntry ("No",  3);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Show axis", idPMMenu);
  
  idPMMenu= glutCreateMenu (menu);
  glutAddMenuEntry ("Yes", 6);
  glutAddMenuEntry ("No",  7);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Show floor XZ", idPMMenu);

  idPMMenu= glutCreateMenu (menu);
  glutAddMenuEntry ("Smooth", 0);
  glutAddMenuEntry ("Flat",   1);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Shade Model", idPMMenu);

  idPMMenu= glutCreateMenu (menu);
  glutAddMenuEntry ("Fill", 4);
  glutAddMenuEntry ("Line", 5);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Polygon Mode", idPMMenu);
  
  idPMMenu= glutCreateMenu (menu);
  glutAddMenuEntry ("Blue", 11);
  glutAddMenuEntry ("Black",12);
  glutAddMenuEntry ("White",13);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Background Color", idPMMenu);
}

/*--------------------------------------------------------------------------
   Mouse...
  --------------------------------------------------------------------------*/
void mouse (int button, int state, int x, int y) {
 
  switch (button) {
  case GLUT_LEFT_BUTTON: {
    if (state== GLUT_DOWN)
      if (glutGetModifiers()== 1)
	mouseButtonPressed= GLUT_LEFT_BUTTON_AND_SHIFT;
      else if (glutGetModifiers()== 2)
	mouseButtonPressed= GLUT_LEFT_BUTTON_AND_CONTROL;
      else if (glutGetModifiers() == 3) // CTRL + SHIFT
	selectObject (x,y);
      else 
	mouseButtonPressed= GLUT_LEFT_BUTTON;
    else
      mouseButtonPressed= NO_BUTTON_PRESSED;
    break;
  }
  }
  glutPostRedisplay ();
} 

/*--------------------------------------------------------------------------
   Mouse motion
  --------------------------------------------------------------------------*/
void motion (int newMouseX, int newMouseY) {


  switch (mouseButtonPressed) {
  case GLUT_LEFT_BUTTON: {  //rotation
    mouseButtonPressed= LEFT_BUTTON_HOLDED;
    oldMouseX= newMouseX; 
    oldMouseY= newMouseY;
    break;
  }
  case GLUT_LEFT_BUTTON_AND_SHIFT: { // translation
    mouseButtonPressed= LEFT_BUTTON_AND_SHIFT_HOLDED;
    oldMouseX= newMouseX; 
    oldMouseY= newMouseY;
    break;
  }
  case GLUT_LEFT_BUTTON_AND_CONTROL: { // zoom
    mouseButtonPressed= LEFT_BUTTON_AND_CONTROL_HOLDED;
    oldMouseX= newMouseX; 
    oldMouseY= newMouseY;
    break;
  }
  case LEFT_BUTTON_HOLDED: {  // rotation
    alfa-= (newMouseX - oldMouseX)/8.0;
    beta+= (newMouseY - oldMouseY)/8.0;
    if (beta<-88) beta= -88.0;
    if (beta> 88) beta=  88.0;
    oldMouseX= newMouseX; 
    oldMouseY= newMouseY;
    break;
  }
  case LEFT_BUTTON_AND_CONTROL_HOLDED: {  // zoom
    radio-= (newMouseY - oldMouseY)/8.0;
    if (radio< 2) radio= 2;
    oldMouseX= newMouseX; 
    oldMouseY= newMouseY;
    break;
  }
  case LEFT_BUTTON_AND_SHIFT_HOLDED: { // translation
    matrixTranslation[0]+= (newMouseX- oldMouseX)/8.0;
    matrixTranslation[1]-= (newMouseY- oldMouseY)/8.0;
    oldMouseX= newMouseX;
    oldMouseY= newMouseY;
    break;
  }
  }
  glutPostRedisplay ();
}

/*--------------------------------------------------------------------------
   Proyection...
  --------------------------------------------------------------------------*/
void proyection (void) {

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective (30.0, aspect, 0.05, 1000.0);
}

/*--------------------------------------------------------------------------
   Reshape...
  --------------------------------------------------------------------------*/
void reshape (int width, int height) {
  
  //GLUI_Master.auto_set_viewport();
  glViewport (0, 0, width, height);
  aspect= (GLdouble) width/ (GLdouble) height;
  proyection();
}

/*--------------------------------------------------------------------------
   Keyboard...
  --------------------------------------------------------------------------*/
void keyboard (unsigned char key, int px, int py) {

  switch (key) {
  case  27: exit (0); break;
  case 'q': exit (0); break;
  case 'Q': exit (0); break;
  }
}

/*--------------------------------------------------------------------------
   Init and create the main window
  --------------------------------------------------------------------------*/
void initWindow (int argc, char **argv) {

  glutInit(&argc, argv);
  glutInitDisplayMode    (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize     (mainWindowWidth, mainWindowHeight);
  glutInitWindowPosition (mainWindowPosX, mainWindowPosY);
  
  main_win = glutCreateWindow ("3DGT - 3D Geometrical Transformations");

  glutDisplayFunc (display);
}

/*--------------------------------------------------------------------------
   Main program
  --------------------------------------------------------------------------*/
int main (int argc, char **argv) {

  initWindow (argc, argv);

  // events...
  GLUI_Master.set_glutReshapeFunc  (reshape);
  GLUI_Master.set_glutKeyboardFunc (keyboard);
  GLUI_Master.set_glutMouseFunc    (mouse); 
  GLUI_Master.set_glutIdleFunc     (NULL);     
  glutMotionFunc                   (motion);

  // inicializations...
  initView ();                    // view...
  initRenderOptions ();           // render options...
  initMenu ();                    // user interface: the menu
  initInterfacePrimitives ();     // user interface: geometric primitives
  initInterfaceTransformations(); // user interface: transformations
  glutMainLoop();                 // the main loop
  return 0;
}

/*--------------------------------------------------------------------------
   End of main.cpp
  --------------------------------------------------------------------------*/
