/*--------------------------------------------------------------------------

   object.cpp v3.0 7/March/2008
   Copyright (C) 2008 J. Ribelles

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

  --------------------------------------------------------------------------*/

#include <GL/glut.h>
#include <stdio.h>
#include <list>

using namespace std;

#include <list>

#include "object.h"
#include "main.h"



//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//------------------ TRANSFORMATION CLASS ----------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------

void Transformation::apply (void) {

  switch (type) {
  case id_translate: glTranslatef (x, y, z);     break;
  case id_rotate:    glRotatef (alpha, x, y, z); break;
  case id_scale:     glScalef (x, y, z);         break;
  }
}

void Transformation::print (FILE *fc) {

  switch (type) {
  case id_translate: fprintf(fc,"    glTranslatef(%.3f,%.3f,%.3f);\n", x, y, z);           break;
  case id_rotate:    fprintf(fc,"    glRotatef (%.3f,%.3f,%.3f,%.3f);\n", alpha, x, y, z); break;
  case id_scale:     fprintf(fc,"    glScalef (%.3f,%.3f,%.3f);\n", x, y, z);              break;
  } 
}

void Transformation::printSave (FILE *fc) {

  fprintf(fc, "2 %d %f %f %f %f\n", type, alpha, x, y, z);
}


//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//------------------ PRIMITIVE CLASS ---------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------

void Primitive::draw (void) {

  switch (type) {
  case id_sphere:
    glutSolidSphere((double)x1,(int)x2,(int)x3);
    break;
  case id_cube:
    glutSolidCube(x1);
    break;
  case id_cone:
    glutSolidCone((double)x1,(double)x2,(int)x3,(int)x4);
    break;
  case id_torus:
    glutSolidTorus((double)x1,(double)x2,(int)x3,(int)x4);
    break;
  case id_dodecahedron:
    glutSolidDodecahedron();
    break;
  case id_octahedron:
    glutSolidOctahedron();
    break;
  case id_tetrahedron:
    glutSolidTetrahedron();
    break;
  case id_icosahedron:
    glutSolidIcosahedron();
    break;
  case id_teapot:
    glutSolidTeapot(x1);
    break;
  }
}

void Primitive::print (FILE *fc, int indent) {

  if (indent) 
    fprintf(fc,"    ");
  else 
    fprintf(fc,"  ");

  switch (type) {
  case id_sphere:
    fprintf(fc,"glutSolidSphere(%.3f,%d,%d);\n", x1, (int)x2, (int)x3);
    break;
  case id_cube:
    fprintf(fc,"glutSolidCube(%.3f);\n", x1);
    break;
  case id_cone:
    fprintf(fc,"glutSolidCone(%.3f,%.3f,%d,%d);\n", x1, x2, (int)x3, (int)x4);
    break;
  case id_torus:
    fprintf(fc,"glutSolidTorus(%.3f,%.3f,%d,%d);\n", x1, x2, (int)x3, (int)x4);
    break;
  case id_dodecahedron:
    fprintf(fc,"glutSolidDodecahedron();\n");
    break;
  case id_octahedron:
    fprintf(fc,"glutSolidOctahedron();\n");
    break;
  case id_tetrahedron:
    fprintf(fc,"glutSolidTetrahedron();\n");
    break;
  case id_icosahedron:
    fprintf(fc,"glutSolidIcosahedron();\n");
    break;
  case id_teapot:
    fprintf(fc,"glutSolidTeapot(%.3f);\n", x1);
    break;
  }
}

void Primitive::printSave (FILE *fc) {

  fprintf (fc,"1 %d %f %f %f %f\n", type, x1, x2, x3, x4);
}



//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//------------------ OBJECT CLASS ------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
int Object::add (int type, float x1, float x2, float x3, float x4) {

  Primitive thePrimitive(cont++, type, x1, x2, x3, x4);
  primitives.push_back (thePrimitive);
  current= --primitives.end();
  return (primitives.size()-1);
}

void Object::draw (void) {

  if (primitives.empty()) return;
  for (list<Primitive>::iterator itera= primitives.begin(); itera!= primitives.end(); itera++) {
    if (itera!= current) 
      glColor3f (1.0,1.0,1.0);
    else 
      glColor3f(1.0,1.0,0.8);
    glPushMatrix();
      itera->applyTransformations();
      itera->draw();
    glPopMatrix();
  }
}

void Object::drawSelect (void) {

  if (primitives.empty()) return;
  for (list<Primitive>::iterator itera= primitives.begin(); itera!= primitives.end(); itera++) {
    glLoadName (itera->getId());
    glPushMatrix();
      itera->applyTransformations();
      itera->draw();
    glPopMatrix();
  }
}

void Object::erase (void) {

  if (primitives.empty()) return;
  primitives.erase(current);
  current= primitives.begin();
}

void Object::setCurrent (int id) {

  list<Primitive>::iterator itera;
  int found= 0;

  if (primitives.empty()) return;
  itera= primitives.begin();
  while ((!found) && (itera!= primitives.end())) {
    if (itera->getId()== id) found= 1;
    else itera++;
  }
  if (found) current= itera;
}

void Object::save (void) {

  FILE *fc;

  if (primitives.empty()) {
    printf ("No data to be saved\n");
    return;
  }

  fc = fopen("data.c","w");
  if (fc== NULL) {
    printf ("File data.c can not be saved\n");
    return;
  }

  fprintf(fc, "/*\n  data.c\n  generated automatically by the application 3GT\n*/\n\n#include <GL/glut.h>\n");
  fprintf(fc,"\nvoid scene (void) {\n");
  for (list<Primitive>::iterator itera= primitives.begin(); itera!= primitives.end(); itera++) {
    fprintf(fc, "\n");
    if (!itera->emptyTransformations()) {
      fprintf(fc, "  glPushMatrix();\n");
      itera->printTransformations(fc);
      itera->print(fc, 1); 
      fprintf(fc, "  glPopMatrix();\n");
    } else
      itera->print(fc, 0);
  }
  fprintf(fc,"}\n");
  fclose (fc);

  fc = fopen("data.save","w");
  if (fc== NULL) {
    printf ("File data.save can not be saved\n");
    return;
  }
  for (list<Primitive>::iterator itera2= primitives.begin(); itera2!= primitives.end(); itera2++) {
    itera2->printSave(fc);
    itera2->printTransformationsSave(fc);
  }
  fclose (fc);
}

bool Object::load (const char *filename) {

  FILE *fc;
  int type, primitive;
  float x, y, z, alpha;
  
  fc = fopen(filename,"r");
  if (fc== NULL) {
    printf ("File %s not found\n", filename);
    return 0;
  }

  while (!primitives.empty()) erase();

  while (!feof(fc)) {
    fscanf(fc,"%d %d %f %f %f %f\n", &type, &primitive, &x, &y, &z, &alpha);
    switch (type) {
    case 1:
      add (primitive, x, y, z, alpha);
      break;
    case 2:
      addTransformation (primitive, y, z, alpha, x);
      break;
    }
  }
  fclose (fc);
  return 1;
}

